from finalsa.sns.client.interface import SnsClient
from pydantic import BaseModel
from typing import Union, Dict, Optional, List


class SnsClientTest(SnsClient):

    def __init__(
        self,
    ) -> None:
        self.topics = {}

    def create_topic(self, name: str):
        self.topics[name] = {
            "arn": f"arn:aws:sns:us-east-1:123456789012:{name}",
            "name": name,
            "id": "123456789012",
            "subscriptions": [],
            "messages": []
        }
        return self.topics[name]

    def subscription_exists(self, topic_name: str, arn: str) -> bool:
        if topic_name in self.topics:
            topic = self.topics[topic_name]
            for sub in topic['subscriptions']:
                if sub['Endpoint'] == arn:
                    return True
        return False

    def get_all_topics(self):
        return list(self.topics.keys())

    def get_or_create_topic(self, name: str):
        if name in self.topics:
            return self.topics[name]
        return self.create_topic(name)

    def get_topic(self, topic_name: str):
        return self.topics.get(topic_name, None)

    def list_subscriptions(self, topic: str):
        if topic in self.topics:
            return self.topics[topic]['subscriptions']
        return []

    def subscribe(self, topic_name: str, protocol: str, endpoint: str) -> Dict:
        if topic_name in self.topics:
            topic = self.topics[topic_name]
            if not self.subscription_exists(topic_name, endpoint):
                topic['subscriptions'].append({
                    "Endpoint": endpoint,
                    "Protocol": protocol
                })
                return {}
        return {}

    def publish(
        self,
        topic_name: str,
        payload: Union[Dict, str],
        attrs: Optional[Dict] = None,
    ) -> Dict:
        "Use this method to publish a message to a topic, it will return an empty dict if the message was published successfully."
        self.get_or_create_topic(topic_name)
        attrs = attrs or {}
        self.topics[topic_name]['messages'].append({
            "Message": payload,
            "Attributes": attrs
        })
        return {}

    def publish_batch(
        self,
        topic_name: str,
        payloads: List[Union[Dict, BaseModel]],
        attrs: Dict | None = ...
    ) -> List[Dict]:
        "Use this method to publish a batch of messages to a topic, it will return an empty dict if the messages were published successfully."
        self.get_or_create_topic(topic_name)
        for payload in payloads:
            self.topics[topic_name]['messages'].append({
                "Message": payload,
                "Attributes": attrs
            })
        return []

    def clear(self):
        self.topics = {}

    def topic_exists(self, topic_name: str) -> bool:
        return topic_name in self.topics

    def messages(self, topic_name: str) -> List[Dict]:
        messages = []
        for message in self.topics[topic_name]['messages']:
            m = message['Message']
            if isinstance(m, str):
                messages.append(m)
            if isinstance(m, dict):
                messages.append(m)
        return messages

    def attributes(self, topic_name: str) -> List[Dict]:
        messages = []
        for message in self.topics[topic_name]['messages']:
            m = message['Attributes']
            if isinstance(m, str):
                messages.append(m)
            if isinstance(m, dict):
                messages.append(m)
        return messages



from finalsa.sns.client.interface import SnsClient
from finalsa.sns.client.client import SnsClientImpl
from finalsa.sns.client.tests import SnsClientTest


__all__ = [
    "SnsClient",
    "SnsClientImpl",
    "SnsClientTest"
]



from finalsa.common.models import to_sqs_message_attributes
from finalsa.sns.client.interface.client import SnsClient
from typing import Union, Dict, List, Optional
from pydantic import BaseModel
import logging
import boto3


class SnsClientImpl(SnsClient):

    def __init__(
        self
    ) -> None:
        self.resource = None
        self.client = None
        self.logger = logging.getLogger("finalsa.clients")
        self.cache = {}

    def get_resource(self):
        if self.resource is None:
            self.resource = boto3.resource('sns')
        return self.resource

    def get_client(self):
        if self.client is None:
            self.client = boto3.client('sns')
        return self.client

    def create_topic(self, name: str):
        try:
            topic = self.get_resource().create_topic(Name=name)
            self.logger.info("Created topic %s with ARN %s.", name, topic.arn)
        except Exception as ex:
            self.logger.error("Couldn't create topic %s.", name)
            self.logger.exception(ex)
            raise ex
        return topic

    def subscription_exists(
        self,
        topic_name: str,
        arn: str
    ) -> bool:
        try:
            subscriptions = self.list_subscriptions(topic_name)
            for sub in subscriptions:
                if sub['Endpoint'] == arn:
                    return True
        except Exception as e:
            self.logger.error("Subscription doesn't exist.")
            self.logger.exception(e)
        return False

    def get_all_topics(self) -> List:
        try:
            topics_iter = self.get_resource().topics.all()
            self.logger.debug("Got topics.")
            return topics_iter
        except Exception as ex:
            self.logger.error("Couldn't get topics.")
            self.logger.exception(ex)
            raise ex

    def get_or_create_topic(self, name: str):
        try:
            topic = self.get_topic(name)
        except Exception as e:
            self.logger.error(
                "Couldn't get topic %s. Creating it.", name)
            self.logger.exception(e)
            topic = self.create_topic(name)
        return topic

    def get_topic(self, topic_name: str):
        if topic_name in self.cache:
            return self.cache[topic_name]
        topic = None
        try:
            topics_iter = self.get_all_topics()
            topics_iter = filter(
                lambda t: t.arn.endswith(topic_name), topics_iter)
            topics = list(topics_iter)
            if len(topics) == 0:
                raise Exception("Topic not found.")
            self.logger.debug("Got topic.")
            topic = topics[0]
            self.cache[topic_name] = topic
        except Exception as e:
            self.logger.error("Couldn't get topics.")
            self.logger.exception(e)
            raise e
        else:
            return topic

    def list_subscriptions(self, topic: str) -> List:
        try:
            topic = self.get_topic(topic)
            return self.get_client().list_subscriptions_by_topic(
                TopicArn=topic.arn)['Subscriptions']
        except Exception as ex:
            self.logger.error("Couldn't get subscriptions.")
            self.logger.exception(ex)
            raise ex

    def subscribe(self, topic_name: str, protocol: str, endpoint: str) -> Dict:
        topic = self.get_topic(topic_name)
        try:
            subscription = self.get_client().subscribe(
                TopicArn=topic.arn,
                Protocol=protocol, Endpoint=endpoint)
            self.logger.info(
                "Subscribed %s to topic %s with protocol %s.",
                endpoint, topic.arn, protocol)
        except Exception as e:
            self.logger.error(
                "Couldn't subscribe %s to topic %s with protocol %s.",
                endpoint, topic.arn, protocol)
            self.logger.exception(e)
            raise e
        else:
            return subscription

    def publish_batch(self, topic_name: str, payloads: List[Union[Dict, BaseModel]], attrs: Dict | None = ...) -> List[Dict]:
        topic = self.get_topic(topic_name)
        attrs = attrs or {}
        entries = self.__dump_batch_payload__(payloads, to_sqs_message_attributes(attrs))
        groups = []
        counter = 0
        helper = []
        for entry in entries:
            if counter == 10:
                groups.append(helper.copy())
                helper.clear()
                counter = 0
            helper.append(entry)
            counter += 1
        if len(helper) > 0:
            groups.append(helper.copy())
        responses = []
        for group in groups:
            try:
                response = self.get_client().publish_batch(
                    TopicArn=topic.arn,
                    PublishBatchRequestEntries=group
                )
                responses.append(response)
            except Exception as e:
                self.logger.error("Failed to publish batch for group: %s", group)
                self.logger.exception(e)
                responses.append({'Error': str(e), 'Group': group})
        return responses

    def publish(
        self,
        topic_name: str,
        payload: Union[Dict, str],
        attrs: Optional[Dict] = None
    ) -> Dict:
        """
        Publish a message to a topic.
        """
        attrs = attrs or {}
        topic = self.get_topic(topic_name)
        if isinstance(payload, dict):
            payload = SnsClientImpl.__dump_payload__(payload)
        response = topic.publish(
            Message=payload,
            MessageAttributes=to_sqs_message_attributes(attrs)
        )
        message_id = response['MessageId']
        self.logger.info(
            "Published message to topic %s.", topic.arn)
        return {
            'id': message_id,
            'topic': topic.arn,
            'payload': payload,
        }



from pydantic import BaseModel
from orjson import dumps
from abc import ABC, abstractmethod
from typing import Union, Dict, List, Optional
from finalsa.traceability import (
    get_correlation_id, get_trace_id, get_span_id
)
from finalsa.traceability.functions import (
    ASYNC_CONTEXT_CORRELATION_ID,
    ASYNC_CONTEXT_TRACE_ID,
    ASYNC_CONTEXT_SPAN_ID,
)
from datetime import datetime, timezone

TOPIC_NAME = 'topic'
SUBTOPIC_NAME = 'subtopic'
PRODUCED_AT = 'produced_at'


class SnsClient(ABC):

    @abstractmethod
    def create_topic(self, name: str):
        pass

    @abstractmethod
    def subscription_exists(self, topic_name: str, arn: str) -> bool:
        pass

    @abstractmethod
    def get_all_topics(self) -> List:
        pass

    @abstractmethod
    def get_or_create_topic(self, name: str):
        pass

    @abstractmethod
    def get_topic(self, topic_name: str):
        pass

    @abstractmethod
    def list_subscriptions(self, topic: str) -> List:
        pass

    @abstractmethod
    def subscribe(self, topic_name: str, protocol: str, endpoint: str) -> Dict:
        pass

    @abstractmethod
    def publish_batch(
        self,
        topic_name: str,
        payloads: List[Union[Dict, BaseModel]],
        attrs: Dict | None = ...
    ) -> List[Dict]:
        pass

    @abstractmethod
    def publish(
        self,
        topic_name: str,
        payload: Union[Dict, str],
        attrs: Optional[Dict] = {}
    ) -> Dict:
        pass

    @staticmethod
    def __dump_payload__(payload: Union[Dict, BaseModel, str]) -> str:
        body = None
        if isinstance(payload, str):
            return payload
        if isinstance(payload, dict):
            body = dumps(payload)
            return body.decode()
        body = payload.model_dump_json()
        return body

    @staticmethod
    def __dump_batch_payload__(payload_list: List[Union[Dict, BaseModel]], attrs:  Dict | None = ...) -> List[Dict[str, str]]:
        result = []
        for payload in payload_list:
            body = SnsClient.__dump_payload__(payload)
            result.append({
                'Message': body,
                'MessageAttributes': attrs
            })
        return result

    @staticmethod
    def __parse_to_message__(
        payload: Union[Dict, BaseModel],
    ) -> Dict:
        if isinstance(payload, BaseModel):
            return payload.model_dump()
        return payload

    def get_default_attrs(
        self,
        topic: Optional[str] = None,
        subtopic: Optional[str] = None,
    ) -> Dict:
        result = {
            ASYNC_CONTEXT_CORRELATION_ID: get_correlation_id(),
            ASYNC_CONTEXT_TRACE_ID: get_span_id(),
            ASYNC_CONTEXT_SPAN_ID: get_trace_id(),
            PRODUCED_AT: datetime.now(timezone.utc).isoformat(),
        }
        if topic:
            result[TOPIC_NAME] = topic
        if subtopic:
            result[SUBTOPIC_NAME] = subtopic
        return result

    def publish_message(
        self,
        topic_name: str,
        payload: Union[Dict, BaseModel],
        subtopic: Optional[str] = None,
    ) -> Dict:
        message_attrs = self.get_default_attrs(topic_name, subtopic)
        message = self.__parse_to_message__(payload)
        return self.publish(topic_name, message, message_attrs)

    def publish_messages_batch(
        self,
        topic_name: str,
        payloads: List[Union[Dict, BaseModel]],
        subtopic: Optional[str] = None,
    ) -> List[Dict]:
        message_attrs = self.get_default_attrs(topic_name, subtopic)
        return self.publish_batch(topic_name, payloads, message_attrs)
